package solutions.bitbadger.documents.core.tests.java.integration;

import solutions.bitbadger.documents.*;
import solutions.bitbadger.documents.core.tests.integration.ThrowawayDatabase;

import java.util.List;
import java.util.Map;

import static org.junit.jupiter.api.Assertions.*;
import static solutions.bitbadger.documents.core.tests.TypesKt.TEST_TABLE;
import static solutions.bitbadger.documents.java.extensions.ConnExt.*;

/**
 * Tests for the JSON-returning functions
 * <p>
 * NOTE: PostgreSQL JSONB columns do not preserve the original JSON with which a document was stored. These tests are
 * the most complex within the library, as they have split testing based on the backing data store. The PostgreSQL tests
 * check IDs (and, in the case of ordered queries, which ones occur before which others) vs. the entire JSON string.
 * Meanwhile, SQLite stores JSON as text, and will return exactly the JSON it was given when it was originally written.
 * These tests can ensure the expected round-trip of the entire JSON string.
 */
final public class JsonFunctions {

    /**
     * PostgreSQL, when returning JSONB as a string, has spaces after commas and colons delineating fields and values.
     * This function will do a crude string replacement to match the target string based on the dialect being tested.
     *
     * @param json The JSON which should be returned
     * @return The actual expected JSON based on the database being tested
     */
    public static String maybeJsonB(String json) throws DocumentException {
        return switch (Configuration.dialect()) {
            case SQLITE -> json;
            case POSTGRESQL -> json.replace("\":", "\": ").replace(",\"", ", \"");
        };
    }

    /**
     * Create a snippet of JSON to find a document ID
     *
     * @param id The ID of the document
     * @return A connection-aware ID to check for presence and positioning
     */
    private static String docId(String id) throws DocumentException {
        return maybeJsonB(String.format("{\"id\":\"%s\"", id));
    }

    public static void allDefault(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonAll(db.getConn(), TEST_TABLE);
        assertTrue(json.startsWith("["), "JSON should start with '[' ($json)");
        switch (Configuration.dialect()) {
            case SQLITE:
                assertTrue(json.contains(JsonDocument.one),
                        String.format("Document 'one' not found in JSON (%s)", json));
                assertTrue(json.contains(JsonDocument.two),
                        String.format("Document 'two' not found in JSON (%s)", json));
                assertTrue(json.contains(JsonDocument.three),
                        String.format("Document 'three' not found in JSON (%s)", json));
                assertTrue(json.contains(JsonDocument.four),
                        String.format("Document 'four' not found in JSON (%s)", json));
                assertTrue(json.contains(JsonDocument.five),
                        String.format("Document 'five' not found in JSON (%s)", json));
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("one")), String.format("Document 'one' not found in JSON (%s)", json));
                assertTrue(json.contains(docId("two")), String.format("Document 'two' not found in JSON (%s)", json));
                assertTrue(json.contains(docId("three")),
                        String.format("Document 'three' not found in JSON (%s)", json));
                assertTrue(json.contains(docId("four")), String.format("Document 'four' not found in JSON (%s)", json));
                assertTrue(json.contains(docId("five")), String.format("Document 'five' not found in JSON (%s)", json));
                break;
        }
        assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)");
    }

    public static void allEmpty(ThrowawayDatabase db) throws DocumentException {
        assertEquals("[]", jsonAll(db.getConn(), TEST_TABLE), "There should have been no documents returned");
    }

    public static void byIdString(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonById(db.getConn(), TEST_TABLE, "two");
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.two, json, "An incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("two")), String.format("An incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void byIdNumber(ThrowawayDatabase db) throws DocumentException {
        Configuration.idField = "key";
        try {
            insert(db.getConn(), TEST_TABLE, new NumIdDocument(18, "howdy"));
            assertEquals(maybeJsonB("{\"key\":18,\"text\":\"howdy\"}"), jsonById(db.getConn(), TEST_TABLE, 18),
                    "The document should have been found by numeric ID");
        } finally {
            Configuration.idField = "id";
        }
    }

    public static void byIdNotFound(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals("{}", jsonById(db.getConn(), TEST_TABLE, "x"), "There should have been no document returned");
    }

    public static void byFieldsMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonByFields(db.getConn(), TEST_TABLE,
                List.of(Field.any("value", List.of("blue", "purple")), Field.exists("sub")), FieldMatch.ALL);
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(String.format("[%s]", JsonDocument.four), json, "The incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
                assertTrue(json.contains(docId("four")),
                        String.format("The incorrect document was returned (%s)", json));
                assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
                break;
        }
    }

    public static void byFieldsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonByFields(db.getConn(), TEST_TABLE, List.of(Field.equal("value", "purple")), null,
                List.of(Field.named("id")));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(String.format("[%s,%s]", JsonDocument.five, JsonDocument.four), json,
                        "The documents were not ordered correctly");
                break;
            case POSTGRESQL:
                final int fiveIdx = json.indexOf(docId("five"));
                final int fourIdx = json.indexOf(docId("four"));
                assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
                assertTrue(fiveIdx >= 0, String.format("Document 'five' not found (%s)", json));
                assertTrue(fourIdx >= 0, String.format("Document 'four' not found (%s)", json));
                assertTrue(fiveIdx < fourIdx,
                        String.format("Document 'five' should have been before 'four' (%s)", json));
                assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
                break;
        }
    }

    public static void byFieldsMatchNumIn(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonByFields(db.getConn(), TEST_TABLE, List.of(Field.any("numValue", List.of(2, 4, 6, 8))));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(String.format("[%s]", JsonDocument.three), json, "The incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
                assertTrue(json.contains(docId("three")),
                        String.format("The incorrect document was returned (%s)", json));
                assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
                break;
        }
    }

    public static void byFieldsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals("[]", jsonByFields(db.getConn(), TEST_TABLE, List.of(Field.greater("numValue", 100))),
                "There should have been no documents returned");
    }

    public static void byFieldsMatchInArray(ThrowawayDatabase db) throws DocumentException {
        for (ArrayDocument doc : ArrayDocument.testDocuments) { insert(db.getConn(), TEST_TABLE, doc); }
        final String json = jsonByFields(db.getConn(), TEST_TABLE, List.of(Field.inArray("values", TEST_TABLE,
                List.of("c"))));
        assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
        assertTrue(json.contains(docId("first")), String.format("The 'first' document was not found (%s)", json));
        assertTrue(json.contains(docId("second")), String.format("The 'second' document was not found (%s)", json));
        assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
    }

    public static void byFieldsNoMatchInArray(ThrowawayDatabase db) throws DocumentException {
        for (ArrayDocument doc : ArrayDocument.testDocuments) { insert(db.getConn(), TEST_TABLE, doc); }
        assertEquals("[]",
                jsonByFields(db.getConn(), TEST_TABLE, List.of(Field.inArray("values", TEST_TABLE, List.of("j")))),
                "There should have been no documents returned");
    }

    public static void byContainsMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonByContains(db.getConn(), TEST_TABLE, Map.of("value", "purple"));
        assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertTrue(json.contains(JsonDocument.four), String.format("Document 'four' not found (%s)", json));
                assertTrue(json.contains(JsonDocument.five), String.format("Document 'five' not found (%s)", json));
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("four")), String.format("Document 'four' not found (%s)", json));
                assertTrue(json.contains(docId("five")), String.format("Document 'five' not found (%s)", json));
                break;
        }
        assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
    }

    public static void byContainsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonByContains(db.getConn(), TEST_TABLE, Map.of("sub", Map.of("foo", "green")),
                List.of(Field.named("value")));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(String.format("[%s,%s]", JsonDocument.two, JsonDocument.four), json,
                        "The documents were not ordered correctly");
                break;
            case POSTGRESQL:
                final int twoIdx = json.indexOf(docId("two"));
                final int fourIdx = json.indexOf(docId("four"));
                assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
                assertTrue(twoIdx >= 0, String.format("Document 'two' not found (%s)", json));
                assertTrue(fourIdx >= 0, String.format("Document 'four' not found (%s)", json));
                assertTrue(twoIdx < fourIdx, String.format("Document 'two' should have been before 'four' (%s)", json));
                assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
                break;
        }
    }

    public static void byContainsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals("[]", jsonByContains(db.getConn(), TEST_TABLE, Map.of("value", "indigo")),
                "There should have been no documents returned");
    }

    public static void byJsonPathMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 10)");
        assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertTrue(json.contains(JsonDocument.four), String.format("Document 'four' not found (%s)", json));
                assertTrue(json.contains(JsonDocument.five), String.format("Document 'five' not found (%s)", json));
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("four")), String.format("Document 'four' not found (%s)", json));
                assertTrue(json.contains(docId("five")), String.format("Document 'five' not found (%s)", json));
                break;
        }
        assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
    }

    public static void byJsonPathMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 10)",
                List.of(Field.named("id")));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(String.format("[%s,%s]", JsonDocument.five, JsonDocument.four), json,
                        "The documents were not ordered correctly");
                break;
            case POSTGRESQL:
                final int fiveIdx = json.indexOf(docId("five"));
                final int fourIdx = json.indexOf(docId("four"));
                assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
                assertTrue(fiveIdx >= 0, String.format("Document 'five' not found (%s)", json));
                assertTrue(fourIdx >= 0, String.format("Document 'four' not found (%s)", json));
                assertTrue(fiveIdx < fourIdx,
                        String.format("Document 'five' should have been before 'four' (%s)", json));
                assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
                break;
        }
    }

    public static void byJsonPathNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals("[]", jsonByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 100)"),
                "There should have been no documents returned");
    }

    public static void firstByFieldsMatchOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonFirstByFields(db.getConn(), TEST_TABLE, List.of(Field.equal("value", "another")));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.two, json, "The incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("two")),
                        String.format("The incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void firstByFieldsMatchMany(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonFirstByFields(db.getConn(), TEST_TABLE, List.of(Field.equal("sub.foo", "green")));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertTrue(json.contains(JsonDocument.two) || json.contains(JsonDocument.four),
                        String.format("Expected document 'two' or 'four' (%s)", json));
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("two")) || json.contains(docId("four")),
                        String.format("Expected document 'two' or 'four' (%s)", json));
                break;
        }
    }

    public static void firstByFieldsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonFirstByFields(db.getConn(), TEST_TABLE, List.of(Field.equal("sub.foo", "green")), null,
                List.of(Field.named("n:numValue DESC")));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.four, json, "An incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("four")),
                        String.format("An incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void firstByFieldsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals("{}", jsonFirstByFields(db.getConn(), TEST_TABLE, List.of(Field.equal("value", "absent"))),
                "There should have been no document returned");
    }

    public static void firstByContainsMatchOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonFirstByContains(db.getConn(), TEST_TABLE, Map.of("value", "FIRST!"));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.one, json, "An incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("one")), String.format("An incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void firstByContainsMatchMany(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonFirstByContains(db.getConn(), TEST_TABLE, Map.of("value", "purple"));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertTrue(json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
                        String.format("Expected document 'four' or 'five' (%s)", json));
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("four")) || json.contains(docId("five")),
                        String.format("Expected document 'four' or 'five' (%s)", json));
                break;
        }
    }

    public static void firstByContainsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonFirstByContains(db.getConn(), TEST_TABLE, Map.of("value", "purple"),
                List.of(Field.named("sub.bar NULLS FIRST")));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.five, json, "An incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("five")),
                        String.format("An incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void firstByContainsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals("{}", jsonFirstByContains(db.getConn(), TEST_TABLE, Map.of("value", "indigo")),
                "There should have been no document returned");
    }

    public static void firstByJsonPathMatchOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonFirstByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ == 10)");
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.two, json, "An incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("two")), String.format("An incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void firstByJsonPathMatchMany(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonFirstByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 10)");
        switch (Configuration.dialect()) {
            case SQLITE:
                assertTrue(json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
                        String.format("Expected document 'four' or 'five' (%s)", json));
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("four")) || json.contains(docId("five")),
                        String.format("Expected document 'four' or 'five' (%s)", json));
                break;
        }
    }

    public static void firstByJsonPathMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String json = jsonFirstByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 10)",
                List.of(Field.named("id DESC")));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.four, json, "An incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("four")),
                        String.format("An incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void firstByJsonPathNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals("{}", jsonFirstByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 100)"),
                "There should have been no document returned");
    }
}
