package solutions.bitbadger.documents.groovy.tests.integration

import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.Dialect
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldMatch

import static org.junit.jupiter.api.Assertions.*
import static solutions.bitbadger.documents.groovy.tests.Types.TEST_TABLE

/**
 * Tests for the JSON-returning functions
 *
 * NOTE: PostgreSQL JSONB columns do not preserve the original JSON with which a document was stored. These tests are
 * the most complex within the library, as they have split testing based on the backing data store. The PostgreSQL tests
 * check IDs (and, in the case of ordered queries, which ones occur before which others) vs. the entire JSON string.
 * Meanwhile, SQLite stores JSON as text, and will return exactly the JSON it was given when it was originally written.
 * These tests can ensure the expected round-trip of the entire JSON string.
 */
final class JsonFunctions {

    /**
     * PostgreSQL, when returning JSONB as a string, has spaces after commas and colons delineating fields and values.
     * This function will do a crude string replacement to match the target string based on the dialect being tested.
     *
     * @param json The JSON which should be returned
     * @return The actual expected JSON based on the database being tested
     */
    static String maybeJsonB(String json) {
        return switch (Configuration.dialect()) {
            case Dialect.SQLITE -> json
            case Dialect.POSTGRESQL -> json.replace ('":', '": ' ).replace (',"', ', "')
        }
    }

    /**
     * Create a snippet of JSON to find a document ID
     *
     * @param id The ID of the document
     * @return A connection-aware ID to check for presence and positioning
     */
    private static String docId(String id) {
        return maybeJsonB("{\"id\":\"$id\"")
    }

    static void allDefault(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonAll(TEST_TABLE)
        assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertTrue(json.contains(JsonDocument.one), "Document 'one' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.two), "Document 'two' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.three), "Document 'three' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.four), "Document 'four' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.five), "Document 'five' not found in JSON ($json)")
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('one')), "Document 'one' not found in JSON ($json)")
                assertTrue(json.contains(docId('two')), "Document 'two' not found in JSON ($json)")
                assertTrue(json.contains(docId('three')), "Document 'three' not found in JSON ($json)")
                assertTrue(json.contains(docId('four')), "Document 'four' not found in JSON ($json)")
                assertTrue(json.contains(docId('five')), "Document 'five' not found in JSON ($json)")
                break
        }
        assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
    }

    static void allEmpty(ThrowawayDatabase db) {
        assertEquals('[]', db.conn.jsonAll(TEST_TABLE), 'There should have been no documents returned')
    }

    static void byIdString(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonById(TEST_TABLE, 'two')
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.two, json, 'An incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('two')), "An incorrect document was returned ($json)")
                break
        }
    }

    static void byIdNumber(ThrowawayDatabase db) {
        Configuration.idField = 'key'
        try {
            db.conn.insert(TEST_TABLE, new NumIdDocument(18, 'howdy'))
            assertEquals(maybeJsonB('{"key":18,"text":"howdy"}'), db.conn.jsonById(TEST_TABLE, 18),
                    'The document should have been found by numeric ID')
        } finally {
            Configuration.idField = 'id'
        }
    }

    static void byIdNotFound(ThrowawayDatabase db) {
        JsonDocument.load(db)
        assertEquals('{}', db.conn.jsonById(TEST_TABLE, 'x'), 'There should have been no document returned')
    }

    static void byFieldsMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonByFields(TEST_TABLE, List.of(Field.any('value', List.of('blue', 'purple')),
                Field.exists('sub')), FieldMatch.ALL)
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals("[${JsonDocument.four}]".toString(), json, 'The incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
                assertTrue(json.contains(docId('four')), "The incorrect document was returned ($json)")
                assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
                break
        }
    }

    static void byFieldsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonByFields(TEST_TABLE, List.of(Field.equal('value', 'purple')), null,
                List.of(Field.named('id')))
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals("[${JsonDocument.five},${JsonDocument.four}]".toString(), json,
                        'The documents were not ordered correctly')
                break
            case Dialect.POSTGRESQL:
                int fiveIdx = json.indexOf(docId('five'))
                int fourIdx = json.indexOf(docId('four'))
                assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
                assertTrue(fiveIdx >= 0, "Document 'five' not found ($json)")
                assertTrue(fourIdx >= 0, "Document 'four' not found ($json)")
                assertTrue(fiveIdx < fourIdx, "Document 'five' should have been before 'four' ($json)")
                assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
                break
        }
    }

    static void byFieldsMatchNumIn(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonByFields(TEST_TABLE, List.of(Field.any('numValue', List.of(2, 4, 6, 8))))
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals("[${JsonDocument.three}]".toString(), json, 'The incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
                assertTrue(json.contains(docId('three')), "The incorrect document was returned ($json)")
                assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
                break
        }
    }

    static void byFieldsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        assertEquals('[]', db.conn.jsonByFields(TEST_TABLE, List.of(Field.greater('numValue', 100))),
                'There should have been no documents returned')
    }

    static void byFieldsMatchInArray(ThrowawayDatabase db) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        String json = db.conn.jsonByFields(TEST_TABLE, List.of(Field.inArray('values', TEST_TABLE, List.of('c'))))
        assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
        assertTrue(json.contains(docId('first')), "The 'first' document was not found ($json)")
        assertTrue(json.contains(docId('second')), "The 'second' document was not found ($json)")
        assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
    }

    static void byFieldsNoMatchInArray(ThrowawayDatabase db) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        assertEquals('[]', db.conn.jsonByFields(TEST_TABLE, List.of(Field.inArray('values', TEST_TABLE, List.of('j')))),
                'There should have been no documents returned')
    }

    static void byContainsMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonByContains(TEST_TABLE, Map.of('value', 'purple'))
        assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertTrue(json.contains(JsonDocument.four), "Document 'four' not found ($json)")
                assertTrue(json.contains(JsonDocument.five), "Document 'five' not found ($json)")
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('four')), "Document 'four' not found ($json)")
                assertTrue(json.contains(docId('five')), "Document 'five' not found ($json)")
                break
        }
        assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
    }

    static void byContainsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonByContains(TEST_TABLE, Map.of('sub', Map.of('foo', 'green')),
                List.of(Field.named('value')))
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals("[${JsonDocument.two},${JsonDocument.four}]", json,
                        'The documents were not ordered correctly')
                break
            case Dialect.POSTGRESQL:
                int twoIdx = json.indexOf(docId('two'))
                int fourIdx = json.indexOf(docId('four'))
                assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
                assertTrue(twoIdx >= 0, "Document 'two' not found ($json)")
                assertTrue(fourIdx >= 0, "Document 'four' not found ($json)")
                assertTrue(twoIdx < fourIdx, "Document 'two' should have been before 'four' ($json)")
                assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
                break
        }
    }

    static void byContainsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        assertEquals('[]', db.conn.jsonByContains(TEST_TABLE, Map.of('value', 'indigo')),
                'There should have been no documents returned')
    }

    static void byJsonPathMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonByJsonPath(TEST_TABLE, '$.numValue ? (@ > 10)')
        assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertTrue(json.contains(JsonDocument.four), "Document 'four' not found ($json)")
                assertTrue(json.contains(JsonDocument.five), "Document 'five' not found ($json)")
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('four')), "Document 'four' not found ($json)")
                assertTrue(json.contains(docId('five')), "Document 'five' not found ($json)")
                break
        }
        assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
    }

    static void byJsonPathMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonByJsonPath(TEST_TABLE, '$.numValue ? (@ > 10)', List.of(Field.named('id')))
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals("[${JsonDocument.five},${JsonDocument.four}]", json,
                        'The documents were not ordered correctly')
                break
            case Dialect.POSTGRESQL:
                int fiveIdx = json.indexOf(docId('five'))
                int fourIdx = json.indexOf(docId('four'))
                assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
                assertTrue(fiveIdx >= 0, "Document 'five' not found ($json)")
                assertTrue(fourIdx >= 0, "Document 'four' not found ($json)")
                assertTrue(fiveIdx < fourIdx, "Document 'five' should have been before 'four' ($json)")
                assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
                break
        }
    }

    static void byJsonPathNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        assertEquals('[]', db.conn.jsonByJsonPath(TEST_TABLE, '$.numValue ? (@ > 100)'),
                'There should have been no documents returned')
    }

    static void firstByFieldsMatchOne(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonFirstByFields(TEST_TABLE, List.of(Field.equal('value', 'another')))
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.two, json, 'The incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('two')), "The incorrect document was returned ($json)")
                break
        }
    }

    static void firstByFieldsMatchMany(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonFirstByFields(TEST_TABLE, List.of(Field.equal('sub.foo', 'green')))
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertTrue(json.contains(JsonDocument.two) || json.contains(JsonDocument.four),
                    "Expected document 'two' or 'four' ($json)")
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('two')) || json.contains(docId('four')),
                        "Expected document 'two' or 'four' ($json)")
                break
        }
    }

    static void firstByFieldsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonFirstByFields(TEST_TABLE, List.of(Field.equal('sub.foo', 'green')), null,
                List.of(Field.named('n:numValue DESC')))
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.four, json, 'An incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('four')), "An incorrect document was returned ($json)")
                break
        }
    }

    static void firstByFieldsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        assertEquals('{}', db.conn.jsonFirstByFields(TEST_TABLE, List.of(Field.equal('value', 'absent'))),
                'There should have been no document returned')
    }

    static void firstByContainsMatchOne(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonFirstByContains(TEST_TABLE, Map.of('value', 'FIRST!'))
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.one, json, 'An incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('one')), "An incorrect document was returned ($json)")
                break
        }
    }

    static void firstByContainsMatchMany(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonFirstByContains(TEST_TABLE, Map.of('value', 'purple'))
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertTrue(json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
                    "Expected document 'four' or 'five' ($json)")
                break
            case Dialect.POSTGRESQL:
                assertTrue( json.contains(docId('four')) || json.contains(docId('five')),
                    "Expected document 'four' or 'five' ($json)")
                break
        }
    }

    static void firstByContainsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonFirstByContains(TEST_TABLE, Map.of('value', 'purple'),
                List.of(Field.named('sub.bar NULLS FIRST')))
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.five, json, 'An incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('five')), "An incorrect document was returned ($json)")
                break
        }
    }

    static void firstByContainsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        assertEquals('{}', db.conn.jsonFirstByContains(TEST_TABLE, Map.of('value', 'indigo')),
                'There should have been no document returned')
    }

    static void firstByJsonPathMatchOne(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonFirstByJsonPath(TEST_TABLE, '$.numValue ? (@ == 10)')
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.two, json, 'An incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('two')), "An incorrect document was returned ($json)")
                break
        }
    }

    static void firstByJsonPathMatchMany(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonFirstByJsonPath(TEST_TABLE, '$.numValue ? (@ > 10)')
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertTrue(json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
                    "Expected document 'four' or 'five' ($json)")
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('four')) || json.contains(docId('five')),
                    "Expected document 'four' or 'five' ($json)")
                break
        }
    }

    static void firstByJsonPathMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        String json = db.conn.jsonFirstByJsonPath(TEST_TABLE, '$.numValue ? (@ > 10)', List.of(Field.named('id DESC')))
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.four, json, 'An incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('four')), "An incorrect document was returned ($json)")
                break
        }
    }

    static void firstByJsonPathNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        assertEquals('{}', db.conn.jsonFirstByJsonPath(TEST_TABLE, '$.numValue ? (@ > 100)'),
                'There should have been no document returned')
    }
}
